import 'dart:collection';
import 'dart:io';

import 'package:flutter/services.dart';
import 'package:flutter_webrtc/flutter_webrtc.dart';
import 'package:random_string/random_string.dart';
import 'package:videosdk/src/core/room/custom_track_configs.dart';

import '../../videosdk.dart';
import '../utils/constants.dart';

class VideoSDK {
  // Method channel to  communicate with platform
  static const MethodChannel _channel = MethodChannel('videosdk');
  static const EventChannel _eventChannel = EventChannel('videosdk-event');

  // States for MediaDevices
  static final mediaDevices = <MediaDeviceType, List<MediaDeviceInfo>>{};

  // Method to join Room
  static Room createRoom({
    required String roomId,
    required String displayName,
    required String token,
    bool micEnabled = true,
    bool camEnabled = true,
    String participantId = '',
    String maxResolution = 'sd',
    int defaultCameraIndex = 0,
    bool multiStream = true,
    CustomTrack? customCameraVideoTrack,
    CustomTrack? customMicrophoneAudioTrack,
    NotificationInfo notification = const NotificationInfo(
      title: "Video SDK",
      message: "Video SDK is sharing screen in the meeting",
      icon: "notification_share",
    ),
    Mode mode = Mode.CONFERENCE,
  }) {
    //
    participantId = participantId.isNotEmpty ? participantId : randomAlpha(8);

    //
    if (token.isEmpty) {
      //
      throw "Token is empty or invalid.";
    }

    Room room = Room(
        participantId: participantId,
        displayName: displayName,
        maxResolution: maxResolution,
        camEnabled: camEnabled,
        micEnabled: micEnabled,
        meetingId: roomId,
        multiStream: multiStream,
        token: token,
        notification: notification,
        defaultCameraIndex: defaultCameraIndex,
        customCameraVideoTrack: customCameraVideoTrack,
        customMicrophoneAudioTrack: customMicrophoneAudioTrack,
        mode: mode);

    _eventChannel.receiveBroadcastStream().listen((event) {
      if (event == "STOP_SCREENSHARE") {
        room.disableScreenShare();
      } else if (event == "START_SCREENSHARE") {
        room.enableShare(iosPermissionGiven: true);
      }
    });

    //
    return room;
  }

  static void requestIOSScreenSharePermission() {
    _channel.invokeMethod('requestScreenSharePermission');
  }

  static Future<Map<String, dynamic>> getDeviceInfo() async {
    String platform;

    if (Platform.isAndroid) {
      platform = "android";
    } else if (Platform.isIOS) {
      platform = "ios";
    } else {
      platform = "web";
    }

    final deviceInfo = LinkedHashMap<String, dynamic>();

    deviceInfo.addAll(<String, dynamic>{
      "sdkType": 'flutter',
      "sdkVersion": sdkVersion,
      "platform": platform
    });

    var deviceUserAgent = await _channel.invokeMethod('getDeviceInfo');

    if (deviceUserAgent != null) {
      deviceInfo.addAll(<String, dynamic>{"deviceUserAgent": deviceUserAgent});
    }

    return deviceInfo;
  }

  // Method used to load media devices
  static Future<Map<MediaDeviceType, List<MediaDeviceInfo>>>
      loadMediaDevices() async {
    // Load MediaDevices
    final List<MediaDeviceInfo> devices =
        await navigator.mediaDevices.enumerateDevices();

    updateDeviceList(devices);
    navigator.mediaDevices.ondevicechange = (event) => {
          navigator.mediaDevices.enumerateDevices().then((devices) {
            updateDeviceList(devices);
          })
        };

    return mediaDevices;
  }

  static updateDeviceList(List<MediaDeviceInfo> devices) {
    // Clear Devices
    mediaDevices[MediaDeviceType.audioInput] = <MediaDeviceInfo>[];
    mediaDevices[MediaDeviceType.audioOutput] = <MediaDeviceInfo>[];
    mediaDevices[MediaDeviceType.videoInput] = <MediaDeviceInfo>[];

    // Iterate MediaDevice
    for (MediaDeviceInfo device in devices) {
      // Conditional Checking
      switch (device.kind) {
        case 'audioinput':
          mediaDevices[MediaDeviceType.audioInput]!.add(device);
          break;
        case 'audiooutput':
          mediaDevices[MediaDeviceType.audioOutput]!.add(device);
          break;
        case 'videoinput':
          mediaDevices[MediaDeviceType.videoInput]!.add(device);
          break;
        default:
          break;
      }
    }
  }

  static Future<List<MediaDeviceInfo>?> getAudioOutputDevices() async {
    await loadMediaDevices();
    return mediaDevices[MediaDeviceType.audioOutput];
  }

  static createMicrophoneAudioTrack(
      {String? microphoneId,
      CustomAudioTrackConfig encoderConfig =
          CustomAudioTrackConfig.speech_standard,
      Map<String, bool> noiseConfig = const {}}) async {
    Map<String, dynamic>? config = customAudioTrackConfigMap[encoderConfig];
    Map<String, dynamic> mediaConstraints = <String, dynamic>{
      "audio": {
        "mandatory": {
          "googEchoCancellation": noiseConfig["echoCancellation"] ??
              config?["echoCancellation"] as bool,
          "googNoiseSuppression": noiseConfig["noiseSuppression"] ??
              config?["noiseSuppression"] as bool,
          "echoCancellation": noiseConfig["echoCancellation"] ??
              config?["echoCancellation"] as bool,
          "googEchoCancellation2": noiseConfig["echoCancellation"] ??
              config?["echoCancellation"] as bool,
          "googDAEchoCancellation": noiseConfig["echoCancellation"] ??
              config?["echoCancellation"] as bool,
          "googAutoGainControl": noiseConfig["autoGainControl"] ??
              config?["autoGainControl"] as bool,
        },
        'optional': microphoneId != null
            ? [
                {
                  'sourceId': microphoneId,
                }
              ]
            : [],
      },
      // "audio": true,
      "video": false
    };

    MediaStream mediaStream =
        await navigator.mediaDevices.getUserMedia(mediaConstraints);
    CustomTrack customVideoTrack = CustomTrack.audioTrack(
        mediaStream: mediaStream,
        audioEncoderConfig: encoderConfig,
        noiseConfig: noiseConfig);
    return customVideoTrack;
  }

  static createCameraVideoTrack(
      {String? cameraId,
      CustomVideoTrackConfig encoderConfig = CustomVideoTrackConfig.h360p_w640p,
      String facingMode = "user",
      // String optimizationMode = "motion",
      bool multiStream = true}) async {
    Map<String, dynamic> mediaConstraints = <String, dynamic>{
      'audio': false,
      'video': {
        'facingMode': facingMode,
        'width': customVideotrackConfigMap[encoderConfig]!["width"],
        'height': customVideotrackConfigMap[encoderConfig]!["height"],
        'frameRate': customVideotrackConfigMap[encoderConfig]!["frameRate"],
        'optional': cameraId != null
            ? [
                {
                  'sourceId': cameraId,
                }
              ]
            : [],
      },
    };

    MediaStream mediaStream =
        await navigator.mediaDevices.getUserMedia(mediaConstraints);
    CustomTrack customVideoTrack = CustomTrack.videoTrack(
      mediaStream: mediaStream,
      multiStream: multiStream,
      videoEncoderConfig: encoderConfig,
      // optimizationMode: optimizationMode
    );
    return customVideoTrack;
  }
}

enum MediaDeviceType { audioInput, audioOutput, videoInput }
